/*
 * ECOFF binary format.
 *
 * Copyright (C) 2003 Juha Aatrokoski, Timo Lilja,
 *   Leena Salmela, Teemu Takanen, Aleksi Virtanen.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: ecoff.c,v 1.5 2003/05/12 17:20:52 lsalmela Exp $
 *
 */
#include "proc/ecoff.h"
#include "lib/libc.h"

/** @name Ecoff loader.
 *
 * This module contains a function to parse useful information from
 * ecoff headers.
 */

/**
 * Parse useful information from a given ecoff file to the ecoff info
 * structure.
 *
 * @param file The ecoff file
 *
 * @param ecoff Information found in the file is returned in this
 * structure. In case of error this structure may contain arbitrary
 * values.
 *
 * @return 0 on failure, other values indicate success.
 */
int ecoff_parse_header(ecoff_info_t *ecoff, openfile_t file)
{
    ecoff_header_t ecoff_hdr;
    ecoff_section_header_t section_hdr;

    int i;

    /* Read the ecoff header */
    if (vfs_read(file, &ecoff_hdr, sizeof(ecoff_header_t)) != 
        sizeof(ecoff_header_t)) {
        return 0;
    }

    /* Check that the ecoff magic word is correct. */
    if (ecoff_hdr.magic1 != ECOFF_MAGIC) {
        return 0;
    }

    /* Zero the return structure in case some segment is not present
       in the given file */
    memoryset(ecoff, 0, sizeof(*ecoff));

    /* Read the section headers. */
    for(i = 0; i < ecoff_hdr.num_sections; i++) {

        if (vfs_read(file, &section_hdr, sizeof(section_hdr)) !=
            sizeof(section_hdr)) {
            return 0;
        }
        
        /* If name does not start with '.' this is no longer a section
           header */
        if (section_hdr.name[0] != '.')
            break;

        /* Check which section we found. */
        if (!stringcmp(section_hdr.name, ".text")) {
            ecoff->text_size = section_hdr.size;
            ecoff->text_vaddr = section_hdr.virtual_addr;
            ecoff->text_location = section_hdr.data_location;
        }

        if (!stringcmp(section_hdr.name, ".rdata")) {
            ecoff->rdata_size = section_hdr.size;
            ecoff->rdata_vaddr = section_hdr.virtual_addr;
            ecoff->rdata_location = section_hdr.data_location;
        }

        if (!stringcmp(section_hdr.name, ".data")) {
            ecoff->data_size = section_hdr.size;
            ecoff->data_vaddr = section_hdr.virtual_addr;
            ecoff->data_location = section_hdr.data_location;
        }

        if (!stringcmp(section_hdr.name, ".bss")) {
            ecoff->bss_size = section_hdr.size;
            ecoff->bss_vaddr = section_hdr.virtual_addr;
        }

    }

    /* The file is ok if there is a text segment. */
    return (ecoff->text_size > 0);

}

/** @} */
