/*
 * ECOFF binary format.
 *
 * Copyright (C) 2003 Juha Aatrokoski, Timo Lilja,
 *   Leena Salmela, Teemu Takanen, Aleksi Virtanen.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: ecoff.h,v 1.4 2003/05/12 17:20:53 lsalmela Exp $
 *
 */

#ifndef BUENOS_PROC_ECOFF
#define BUENOS_PROC_ECOFF

#include "lib/types.h"
#include "fs/vfs.h"

/* Structure of the ecoff header found in the beginning of a ecoff
 * file. This is actually a concatenation of the ecoff file header and
 * the aout header that follows the file header.
 */
typedef struct {
    /* The ecoff magic word used to identify the file as an ecoff file */
    uint16_t magic1               __attribute__ ((packed));
    /* The number of sections in the object file. */
    uint16_t num_sections         __attribute__ ((packed));
    /* When this file was created. */
    uint32_t timestamp            __attribute__ ((packed));
    /* Location of symbol table in this file. */
    uint32_t sym_table_location   __attribute__ ((packed));
    /* Number of symbol table entries */
    uint32_t num_sym_entries      __attribute__ ((packed));
    /* Size of optional header */
    uint16_t opt_hdr_size         __attribute__ ((packed));
    /* Some flag bits (not used in BUENOS) */
    uint16_t flags                __attribute__ ((packed));
    /* Start of AOUT header */
    /* The AOUT header magic number */
    uint16_t magic2               __attribute__ ((packed));
    /* Version number */
    uint16_t version              __attribute__ ((packed));
    /* Size of text segment */
    uint32_t text_size            __attribute__ ((packed));
    /* Size of data segment */
    uint32_t data_size            __attribute__ ((packed));
    /* Size of bss segment */
    uint32_t bss_size             __attribute__ ((packed));
    /* Where the execution of this binary should be started from */
    uint32_t entry_point          __attribute__ ((packed));
    /* Start (virtual) address of text segment */
    uint32_t text_start           __attribute__ ((packed));
    /* Start (virtual) address of data segment */
    uint32_t data_start           __attribute__ ((packed));
    /* Start (virtual) address of bss segment */
    uint32_t bss_start            __attribute__ ((packed));
    /* Global pointer mask */
    uint32_t gprmask              __attribute__ ((packed));
    /* ??? */
    uint32_t cprmask[4]           __attribute__ ((packed));
    /* Value of global pointer */
    uint32_t gp                   __attribute__ ((packed));
} ecoff_header_t;

/* Structure of section headers in the ecoff format. These come after
 * the ecoff header described above. The field num_sections in the
 * ecoff header define the number of section headers.
 */
typedef struct {
    /* Name of the segment. */
    char name[8]                __attribute__ ((packed));
    /* Physical address of the segment */
    uint32_t phys_addr          __attribute__ ((packed));
    /* Virtual address of the segment */
    uint32_t virtual_addr       __attribute__ ((packed));
    /* Size of the segment */
    uint32_t size               __attribute__ ((packed));
    /* Location in this file of the raw data */
    uint32_t data_location      __attribute__ ((packed));
    /* Location of relocation information */
    uint32_t reloc_location     __attribute__ ((packed));
    /* Location of line number information */
    uint32_t line_num_location  __attribute__ ((packed));
    /* Number of relocation entries */
    uint16_t num_reloc          __attribute__ ((packed));
    /* Number of line number entries */
    uint16_t num_line_nums      __attribute__ ((packed));
    /* Some flags (not used in BUENOS) */
    uint32_t flags              __attribute__ ((packed));
} ecoff_section_header_t;

/* Return data type for the ecoff header parser. This structure
 * contains the information about text, rdata, data and bss segments.
 */
typedef struct {
    uint32_t text_size;      /* Text segment size */
    uint32_t text_vaddr;     /* Virtual address of the text segment */
    uint32_t text_location;  /* Location of text segment in the file */

    uint32_t rdata_size;     /* Rdata segment size */
    uint32_t rdata_vaddr;    /* Virtual address of the rdata segment */
    uint32_t rdata_location; /* Location of rdata segment in the file */

    uint32_t data_size;      /* Data segment size */
    uint32_t data_vaddr;     /* Virtual address of the data segment */
    uint32_t data_location;  /* Location of data segment in the file */

    uint32_t bss_size;       /* Bss segment size */
    uint32_t bss_vaddr;      /* Virtaul address of the bss segment */
} ecoff_info_t;

int ecoff_parse_header(ecoff_info_t *ecoff, openfile_t file);

/* Magic number for the ecoff header. */
#define ECOFF_MAGIC 0x160

#endif
