/*
 * Process startup.
 *
 * Copyright (C) 2003 Juha Aatrokoski, Timo Lilja,
 *   Leena Salmela, Teemu Takanen, Aleksi Virtanen.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: process.c,v 1.7 2003/05/25 12:31:50 ttakanen Exp $
 *
 */

#include "proc/process.h"
#include "proc/ecoff.h"
#include "kernel/thread.h"
#include "kernel/assert.h"
#include "kernel/interrupt.h"
#include "kernel/config.h"
#include "fs/vfs.h"
#include "drivers/yams.h"
#include "vm/vm.h"
#include "vm/pagepool.h"

/** @name Process startup
 *
 * This module contains a function to start a userland process.
 */

/**
 * Starts one userland process. The thread calling this function will
 * be used to run the process and will therefore never return from
 * this function. This function asserts that no errors occur in
 * process startup (the executable file exists and is a valid ecoff
 * file, enough memory is available, file operations succeed...).
 * Therefore this function is not suitable to allow startup of
 * arbitrary processes.
 *
 * @executable The name of the executable to be run in the userland
 * process
 */
void process_start(const char *executable)
{
    thread_table_t *my_entry;
    pagetable_t *pagetable;
    openfile_t file;
    ecoff_info_t ecoff;
    uint32_t phys_page;
    context_t user_context;
    uint32_t stack_bottom;

    int text_pages;
    int rdata_pages;
    int data_pages;
    int bss_pages;

    int i;

    interrupt_status_t intr_status;

    my_entry = thread_get_current_thread_entry();

    /* If the pagetable of this thread is not NULL, we are trying to
       run a userland process for a second time in the same thread.
       This is not possible. */
    KERNEL_ASSERT(my_entry->pagetable == NULL);

    pagetable = vm_create_pagetable(thread_get_current_thread());

    intr_status = _interrupt_disable();
    my_entry->pagetable = pagetable;
    _interrupt_set_state(intr_status);

    file = vfs_open((char *)executable);

    /* Make sure the file existed and that the ecoff loader finds it
       ok. */
    KERNEL_ASSERT(file >= 0);
    KERNEL_ASSERT(ecoff_parse_header(&ecoff, file));

    /* Calculate the number of pages needed for each segment */
    text_pages = ((ecoff.text_size + PAGE_SIZE - 1)/ PAGE_SIZE);
    rdata_pages = ((ecoff.rdata_size + PAGE_SIZE - 1)/ PAGE_SIZE);
    data_pages = ((ecoff.data_size + PAGE_SIZE - 1)/ PAGE_SIZE);
    bss_pages = ((ecoff.bss_size + PAGE_SIZE - 1)/ PAGE_SIZE);

    /* Calculate the number of pages needed by the whole process
       (including userland stack). Since we don't have proper tlb
       handling code, all these pages must fit into TLB. */
    KERNEL_ASSERT(text_pages + rdata_pages + data_pages + bss_pages + 
                  CONFIG_USERLAND_STACK_SIZE <= 
                  (int)_tlb_get_maxindex() + 1);

    /* Allocate and map stack */
    for(i = 0; i < CONFIG_USERLAND_STACK_SIZE; i++) {
        phys_page = pagepool_get_phys_page();
        KERNEL_ASSERT(phys_page != 0);
        vm_map(my_entry->pagetable, phys_page, 
               (USERLAND_STACK_TOP & PAGE_SIZE_MASK) - i*PAGE_SIZE, 1);
    }

    /* Allocate and map pages for the segments. We assume that
       segments begin at page boundary. (The linker script in tests
       directory creates this kind of segments) */
    for(i = 0; i < text_pages; i++) {
        phys_page = pagepool_get_phys_page();
        KERNEL_ASSERT(phys_page != 0);
        vm_map(my_entry->pagetable, phys_page, 
               ecoff.text_vaddr + i*PAGE_SIZE, 1);
    }

    for(i = 0; i < rdata_pages; i++) {
        phys_page = pagepool_get_phys_page();
        KERNEL_ASSERT(phys_page != 0);
        vm_map(my_entry->pagetable, phys_page, 
               ecoff.rdata_vaddr + i*PAGE_SIZE, 1);
    }

    for(i = 0; i < data_pages; i++) {
        phys_page = pagepool_get_phys_page();
        KERNEL_ASSERT(phys_page != 0);
        vm_map(my_entry->pagetable, phys_page, 
               ecoff.data_vaddr + i*PAGE_SIZE, 1);
    }

    for(i = 0; i < bss_pages; i++) {
        phys_page = pagepool_get_phys_page();
        KERNEL_ASSERT(phys_page != 0);
        vm_map(my_entry->pagetable, phys_page, 
               ecoff.bss_vaddr + i*PAGE_SIZE, 1);
    }

    /* Put the mapped pages into TLB. Here we again assume that the
       pages fit into the TLB. After writing proper TLB exception
       handling this call should be skipped. */
    intr_status = _interrupt_disable();
    tlb_fill(my_entry->pagetable);
    _interrupt_set_state(intr_status);
    
    /* Now we may use the virtual addresses of the segments. */

    /* Zero the pages. */
    memoryset((void *)ecoff.text_vaddr, 0, text_pages*PAGE_SIZE);
    memoryset((void *)ecoff.rdata_vaddr, 0, rdata_pages*PAGE_SIZE);
    memoryset((void *)ecoff.data_vaddr, 0, data_pages*PAGE_SIZE);
    memoryset((void *)ecoff.bss_vaddr, 0, bss_pages*PAGE_SIZE);

    stack_bottom = (USERLAND_STACK_TOP & PAGE_SIZE_MASK) - 
        (CONFIG_USERLAND_STACK_SIZE-1)*PAGE_SIZE;
    memoryset((void *)stack_bottom, 0, CONFIG_USERLAND_STACK_SIZE*PAGE_SIZE);

    /* Copy segments */

    /* Make sure that the text segment is in proper place. */
    KERNEL_ASSERT(ecoff.text_vaddr >= PAGE_SIZE);
    KERNEL_ASSERT(vfs_seek(file, ecoff.text_location) == VFS_OK);
    KERNEL_ASSERT(vfs_read(file, (void *)ecoff.text_vaddr, ecoff.text_size) ==
                  (int)ecoff.text_size);

    if (ecoff.rdata_size > 0) {
        KERNEL_ASSERT(ecoff.rdata_vaddr >= PAGE_SIZE);
        KERNEL_ASSERT(vfs_seek(file, ecoff.rdata_location) == VFS_OK);
        KERNEL_ASSERT(vfs_read(file, (void *)ecoff.rdata_vaddr, 
                               ecoff.rdata_size) ==
                      (int)ecoff.rdata_size);
    }

    if (ecoff.data_size > 0) {
        KERNEL_ASSERT(ecoff.data_vaddr >= PAGE_SIZE);
        KERNEL_ASSERT(vfs_seek(file, ecoff.data_location) == VFS_OK);
        KERNEL_ASSERT(vfs_read(file, (void *)ecoff.data_vaddr, 
                               ecoff.data_size) ==
                      (int)ecoff.data_size);
    }

    if (ecoff.bss_size > 0) {
        KERNEL_ASSERT(ecoff.bss_vaddr >= PAGE_SIZE);
    }

    /* Set the dirty bit to zero (read-only) on read-only pages. */
    
    for(i = 0; i < text_pages; i++) {
        vm_set_dirty(my_entry->pagetable, ecoff.text_vaddr + i*PAGE_SIZE, 0);
    }

    for(i = 0; i < rdata_pages; i++) {
        vm_set_dirty(my_entry->pagetable, ecoff.rdata_vaddr + i*PAGE_SIZE, 0);
    }

    /* Initialize the user context. (Status register is handled by
       thread_goto_userland) */
    memoryset(&user_context, 0, sizeof(user_context));
    user_context.cpu_regs[MIPS_REGISTER_SP] = USERLAND_STACK_TOP;
    user_context.pc = ecoff.text_vaddr;

    thread_goto_userland(&user_context);

    KERNEL_PANIC("thread_goto_userland failed.");
}

/** @} */
